/**
 * 
 * AiCiA - Android IRC Client
 * 
 * Copyright (C)2010 GORRY.
 * 
 */

package net.gorry.aicia;

import java.io.IOException;

import org.xmlpull.v1.XmlPullParserException;

import android.content.Context;
import android.content.res.ColorStateList;
import android.content.res.Resources;
import android.content.res.XmlResourceParser;
import android.os.Bundle;
import android.os.Handler;
import android.text.Editable;
import android.text.SpannableStringBuilder;
import android.text.Spanned;
import android.text.method.KeyListener;
import android.text.method.LinkMovementMethod;
import android.text.method.MovementMethod;
import android.text.style.ForegroundColorSpan;
import android.text.util.Linkify;
import android.util.Log;
import android.view.KeyEvent;
import android.view.MotionEvent;
import android.view.View;
import android.view.ViewGroup;
import android.view.ViewGroup.LayoutParams;
import android.widget.LinearLayout;
import android.widget.ScrollView;
import android.widget.TextView;

/**
 * 
 * ログウィンドウ
 * 
 * @author GORRY
 *
 */
public class LogWindow {
	private static final String TAG = "LogWindow";
	private static final boolean V = true;//false;
	@SuppressWarnings("unused")
	private static final int WC = ViewGroup.LayoutParams.WRAP_CONTENT;
	@SuppressWarnings("unused")
	private static final int FP = ViewGroup.LayoutParams.FILL_PARENT;

	private ScrollView mScrollView;
	private LogTextView mTextView;
	private final Handler mHandler = new Handler();
	private String mWindowName;
	private int mTextColorResourceId;
	private int mDateColor;
	private int mTextSize = 0;
	private int mLogBufferSize;
	private int mParentLayoutWidth;
	private int mParentLayoutHeight;
	private LinearLayout mParentLayout;
	private final Context me;


	/**
	 * コンストラクタ
	 * @param context コンテキスト
	 * @param name ウィンドウ名
	 */
	public LogWindow(final Context context, final String name) {
		if (V) Log.v(TAG, "LogWindow()");
		me = context;
		mWindowName = name;
	}

	/**
	 * ログバッファサイズの設定
	 * ログサイズが設定値の２倍になると、設定値までトリミングされる。
	 * @param size ログバッファサイズ
	 */
	public void setLogBufferSize(final int size) {
		if (V) Log.v(TAG, "setLogBufferSize()");
		mLogBufferSize = size;
	}

	/**
	 * TextViewの取得
	 * @return TextView
	 */
	public TextView getTextView() {
		if (V) Log.v(TAG, "getTextView()");
		return mTextView;
	}

	/**
	 * ScrollViewの取得
	 * @return ScrollView
	 */
	public ScrollView getScrollView() {
		if (V) Log.v(TAG, "getScrollView()");
		return mScrollView;
	}

	/**
	 * ステートセーブ
	 * @param outState ステート情報格納先
	 */
	public void saveInstanceState(final Bundle outState) {
		if (V) Log.v(TAG, "saveInstanceState()");
		//		outState.putCharSequence(mWindowName, mTextView.getText());
	}

	/**
	 * ステートロード
	 * @param inState ステート情報格納先
	 */
	public void restoreInstanceState(final Bundle inState) {
		if (V) Log.v(TAG, "restoreInstanceState()");
		//		mTextView.setText(inState.getCharSequence(mWindowName));
	}

	/**
	 * ウィンドウ名の設定
	 * @param name ウィンドウ名
	 */
	public void setWindowName(final String name) {
		if (V) Log.v(TAG, "setWindowName()");
		mWindowName = name;
	}

	/**
	 * ウィンドウ名の取得
	 * @return ウィンドウ名
	 */
	public String getWindowName() {
		if (V) Log.v(TAG, "getWindowName()");
		return mWindowName;
	}

	/**
	 * テキストフォントサイズの設定
	 * @param size フォントサイズ
	 */
	public void setTextSize(final int size) {
		if (V) Log.v(TAG, "setTextSize()");
		mTextSize = size;
		if (mTextView != null) {
			if (mTextSize > 0) {
				mTextView.setTextSize(mTextSize);
			}
		}
	}

	/**
	 * テキストカラーリソースの設定
	 * @param id テキストカラーリソースID
	 */
	public void setTextColorResource(final int id) {
		if (V) Log.v(TAG, "setTextColorResource()");
		mTextColorResourceId = id;
	}

	/**
	 * 日付時刻カラーの設定
	 * @param color Color.rgb()などによるカラー値
	 */
	public void setDateColor(final int color) {
		if (V) Log.v(TAG, "setDateColor()");
		mDateColor = color;
	}

	/**
	 * ウィンドウ横幅の設定
	 * @param w 横幅
	 */
	public void setWidth(final int w) {
		mTextView.setWidth(w);
	}

	/**
	 * 親レイアウトの縦横幅を登録
	 * @param w 横幅
	 * @param h 縦幅
	 */
	public void registerParentLayout(final int w, final int h) {
		mParentLayoutWidth = w;
		mParentLayoutHeight = h;
		mParentLayout.updateViewLayout(
				mScrollView,
				new LinearLayout.LayoutParams(
						LayoutParams.FILL_PARENT,
						LayoutParams.FILL_PARENT,
						3
				)
		);
	}

	/**
	 * フリックでチャンネル移動可能なテキストビュー
	 */
	private class LogTextView extends TextView {
		private int xOnTouch = 0;
		private int yOnTouch = 0;
		private boolean shift = false;

		/**
		 * コンストラクタ
		 * @param context コンテキスト
		 */
		public LogTextView(final Context context) {
			super(context);
		}

		/*
		 * @see android.widget.TextView#onTouchEvent(android.view.MotionEvent)
		 */
		@Override
		public boolean onTouchEvent(final MotionEvent event) {
			final boolean result = super.onTouchEvent(event);
			switch (event.getAction()) {
				case MotionEvent.ACTION_DOWN:
				{
					xOnTouch = (int)event.getX();
					yOnTouch = (int)event.getY();
					shift = false;
					if (V) Log.v(TAG, "ACTION_DOWN: X=" + xOnTouch + ",Y="  + yOnTouch);
					return true;
				}
				case MotionEvent.ACTION_OUTSIDE:
				{
					shift = true;
					if (V) Log.v(TAG, "ACTION_OUTSIDE: X=" + xOnTouch + ",Y="  + yOnTouch);
					return true;
				}
				case MotionEvent.ACTION_UP:
				{
					final int xOnRelease = (int)event.getX();
					final int yOnRelease = (int)event.getY();
					final int xMove = xOnRelease - xOnTouch;
					final int yMove = yOnRelease - yOnTouch;
					final int smallWidth = (mParentLayoutWidth > mParentLayoutHeight) ? mParentLayoutWidth : mParentLayoutHeight;
					final int edge = event.getEdgeFlags();
					if (V) Log.v(TAG, "ACTION_UP: X=" + xOnRelease + ",Y=" + yOnRelease + ",W=" + xMove + ",H=" + yMove + ",EDGE=" + edge);
					if (Math.abs(xMove) > Math.abs(yMove)) {
						if (Math.abs(xMove) >= smallWidth/8) {
							boolean changed = false;
							if (xMove > 0) {
								final int f = (shift ? 1 : 0);
								changed = ActivityMain.doMain.doChangeNextChannel(-1, f, false);
							} else {
								final int f = (shift ? 1 : 0);
								changed = ActivityMain.doMain.doChangeNextChannel(1, f, false);
							}
							if (!changed) {
								ActivityMain.myShortToastShow(me.getString(R.string.activitymain_java_notchanged));
							}
							return true;
						}
					}
					xOnTouch = xOnRelease;
					yOnTouch = yOnRelease;
					shift = false;
					return result;
				}
			}
			return result;
		}

	}

	/**
	 * ログウィンドウを作成
	 * @param context コンテキスト
	 * @param parentLayout 親レイアウト
	 */
	public void createWindow(final Context context, final LinearLayout parentLayout) {
		if (V) Log.v(TAG, "createWindow()");
		mParentLayout = parentLayout;
		mScrollView = new ScrollView(context);
		mScrollView.setLayoutParams(
				new LinearLayout.LayoutParams(
						LayoutParams.FILL_PARENT,
						LayoutParams.FILL_PARENT,
						3
				)
		);
		parentLayout.addView(mScrollView);

		{
			mTextView = new LogTextView(context);
			mTextView.setLayoutParams(
					new LinearLayout.LayoutParams(
							LayoutParams.FILL_PARENT,
							LayoutParams.FILL_PARENT,
							4
					)
			);
			if (mTextSize > 0) {
				mTextView.setTextSize(mTextSize);
			}
			// textView.setAutoLinkMask(Linkify.ALL);
			final MovementMethod movementmethod = LinkMovementMethod.getInstance();
			mTextView.setMovementMethod(movementmethod);
			mTextView.setFocusableInTouchMode(true);

			final Resources resources = me.getResources();
			final XmlResourceParser parser = resources.getXml(mTextColorResourceId);
			ColorStateList colors = null;
			try {
				colors = ColorStateList.createFromXml(resources, parser);
			} catch (final XmlPullParserException e) {
				e.printStackTrace();
			} catch (final IOException e) {
				e.printStackTrace();
			}
			mTextView.setTextColor(colors);

			mScrollView.addView(mTextView);

			mTextView.setKeyListener(onTextViewKey);
			// mScrollView.setOnTouchListener(onTextViewTouch);
		}
	}

	/**
	 * ログウィンドウの破棄
	 */
	public void destroyWindow() {
		if (V) Log.v(TAG, "destroyWindow()");
		if (mParentLayout != null) {
			mParentLayout.removeAllViews();
			mTextView = null;
			mScrollView = null;
		}
	}


	/**
	 * ログウィンドウのキーリスナ処理
	 */
	private final KeyListener onTextViewKey = new KeyListener() {
		public int getInputType() {
			return 0;
		}

		public boolean onKeyDown(final View view, final Editable text, final int keyCode, final KeyEvent event) {
			/*
			if (keyCode == KeyEvent.KEYCODE_ENTER) {
				ActivityMain.layout.mInputBox.requestFocus();
				return true;
			}
			 */
			if (EditAssist.isAnyKey(keyCode)) {
				ActivityMain.layout.mInputBox.requestFocus();
				return true;
			}
			final boolean k = ActivityMain.doMain.doShortcutKey(keyCode, event);
			if (k == true) return true;
			return false;
		}

		public boolean onKeyUp(final View view, final Editable text, final int keyCode, final KeyEvent event) {
			return false;
		}

		public boolean onKeyOther(final View view, final Editable text, final KeyEvent event) {
			return false;
		}

		public void clearMetaKeyState(final View view, final Editable content, final int states) {
			return;
		}

	};

	/**
	 * ログウィンドウのタッチリスナ処理
	 */
	/*
	private final OnTouchListener onTextViewTouch = new OnTouchListener() {
		int xOnTouch = 0;
		int yOnTouch = 0;
		public boolean onTouch(final View v, final MotionEvent event) {
			switch (event.getAction()) {
				case MotionEvent.ACTION_DOWN:
				{
					xOnTouch = (int)event.getX();
					yOnTouch = (int)event.getY();
					if (V) Log.v(TAG, "ACTION_DOWN: X=" + xOnTouch + ",Y="  + yOnTouch);
					return v.onTouchEvent(event);
				}
				case MotionEvent.ACTION_UP:
				{
					final int xOnRelease = (int)event.getX();
					final int yOnRelease = (int)event.getY();
					final int xMove = xOnRelease - xOnTouch;
					final int yMove = yOnRelease - yOnTouch;
					final int smallWidth = (mParentLayoutWidth > mParentLayoutHeight) ? mParentLayoutWidth : mParentLayoutHeight;
					if (V) Log.v(TAG, "ACTION_UP: X=" + xOnRelease + ",Y=" + yOnRelease + ",W=" + xMove + ",H=" + yMove);
					if (Math.abs(xMove) > Math.abs(yMove)) {
						if (Math.abs(xMove) >= smallWidth/4) {
							boolean changed = false;
							if (xMove > 0) {
								changed = ActivityMain.doMain.doChangeNextChannel(1, 0, false);
							} else {
								changed = ActivityMain.doMain.doChangeNextChannel(-1, 0, false);
							}
							if (!changed) {
								ActivityMain.myShortToastShow(me.getString(R.string.activitymain_java_notchanged));
							}
							return true;
						}
					}
					xOnTouch = xOnRelease;
					yOnTouch = yOnRelease;
					return false;
				}
			}
			return false;
		}

	};
	 */

	/**
	 * ログウィンドウにSpanログを登録
	 * @param ssb Spanログ
	 */
	public synchronized void setMessage(final SpannableStringBuilder ssb) {
		if (V) Log.v(TAG, "setMessage()");
		final Runnable scroller = new Runnable() {
			public void run() {
				if (V) Log.v(TAG, "scroller start");
				mScrollView.scrollTo(0, mTextView.getHeight() - mScrollView.getHeight());
				if (V) Log.v(TAG, "scroller end");
			}
		};
		final Runnable setter = new Runnable() {
			public void run() {
				if (V) Log.v(TAG, "setter start");
				mTextView.setText(ssb);
				if (V) Log.v(TAG, "setter end");
				mHandler.post(scroller);
			}
		};
		mHandler.post(setter);
	}

	/**
	 * ログウィンドウにSpanログを追加
	 * @param ssb Spanログ
	 */
	public synchronized void addMessage(final SpannableStringBuilder ssb) {
		if (V) Log.v(TAG, "addMessage()");
		final int tvHeight = mTextView.getHeight();
		final int svHeight = mScrollView.getHeight();
		final int svScrollY = mScrollView.getScrollY();
		@SuppressWarnings("unused")
		final int scrollViewOldY = tvHeight - svHeight;
		final boolean viewOldLog = (tvHeight-svScrollY-svHeight > mParentLayoutHeight);  // １画面以上戻っていたらスクロールしない
		final Runnable scroller = new Runnable() {
			public void run() {
				// 行が増えて値が変わるので取り直し
				if (V) Log.v(TAG, "scroller start");
				mScrollView.smoothScrollTo(0, mTextView.getHeight() - mScrollView.getHeight());
				if (V) Log.v(TAG, "scroller end");
			}
		};
		final Runnable adder = new Runnable() {
			public void run() {
				if (mTextView.length() > mLogBufferSize) {
					// バッファトリミング
					if (V) Log.v(TAG, "get start");
					final CharSequence buf = mTextView.getText();
					if (V) Log.v(TAG, "get end");
					if (V) Log.v(TAG, "trim start");
					int start = buf.length()/2;
					final int end = buf.length();
					for (;start<end-1; start++) {
						if (buf.charAt(start) == '\n') {
							start++;
							break;
						}
					}
					if (V) Log.v(TAG, "trim end");
					if (V) Log.v(TAG, "trim set start");
					mTextView.setText(buf.subSequence(start, end));
					if (V) Log.v(TAG, "trim set end");
				}
				if (V) Log.v(TAG, "adder start");
				mTextView.append(ssb);
				if (V) Log.v(TAG, "adder end");
				if (!viewOldLog) {
					mHandler.post(scroller);
				}
			}
		};
		mHandler.post(adder);
	}

	/**
	 * ログウィンドウにテキストメッセージを追加
	 * @param message テキストメッセージ
	 * @param dateMsg 日付時刻
	 */
	public synchronized void addMessage(final String message, final String dateMsg) {
		if (V) Log.v(TAG, "addMessage()");
		final int tvHeight = mTextView.getHeight();
		final int svHeight = mScrollView.getHeight();
		final int svScrollY = mScrollView.getScrollY();
		@SuppressWarnings("unused")
		final int scrollViewOldY = tvHeight - svHeight;
		final boolean viewOldLog = (tvHeight-svScrollY-svHeight > mParentLayoutHeight);  // １画面以上戻っていたらスクロールしない
		final Runnable scroller = new Runnable() {
			public void run() {
				// 行が増えて値が変わるので取り直し
				mScrollView.smoothScrollTo(0, mTextView.getHeight() - mScrollView.getHeight());
			}
		};
		final Runnable adder = new Runnable() {
			public void run() {
				SpannableStringBuilder span = null;
				String msg = message + "\n";
				switch (SystemConfig.dateMode) {
					default:
					case 0:
					{
						span = new SpannableStringBuilder(msg);
						break;
					}
					case 1:
					{
						msg = dateMsg + " " + message + "\n";
						span = new SpannableStringBuilder(msg);
						final ForegroundColorSpan c = new ForegroundColorSpan(mDateColor);
						span.setSpan(c, 0, 5, Spanned.SPAN_EXCLUSIVE_EXCLUSIVE);
						break;
					}
				}
				Linkify.addLinks(span, Linkify.WEB_URLS);
				if (mTextView.length() > mLogBufferSize) {
					// バッファトリミング
					final CharSequence buf = mTextView.getText();
					int start = buf.length()/2;
					final int end = buf.length();
					for (;start<end-1; start++) {
						if (buf.charAt(start) == '\n') {
							start++;
							break;
						}
					}
					mTextView.setText(buf.subSequence(start, end));
				}
				mTextView.append(span);
				if (!viewOldLog) {
					mHandler.post(scroller);
				}
			}
		};
		mHandler.post(adder);
	}

	/**
	 * ログのクリア
	 */
	public synchronized void clearLog() {
		final Runnable clearer = new Runnable() {
			public void run() {
				mTextView.setText("");
			}
		};
		mHandler.post(clearer);
	}


}
